@@ -1,5 +1,6 @@
 /* This file is part of the KDE libraries
    Copyright (C) 2005 Olivier Goffart <ogoffart at kde.org>
+   Copyright (C) 2013-2015 Martin Klapetek <mklapetek@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -19,31 +20,39 @@
 #include "knotificationmanager_p.h"
 #include "knotification.h"
 
-#include <QDebug>
 #include <QHash>
 #include <QWidget>
 #include <QDBusConnection>
 #include <QPointer>
 #include <QBuffer>
-
-#include <kservicetypetrader.h>
+#include <KPluginLoader>
 
 #include "knotifyconfig.h"
-#include "knotifyplugin.h"
+#include "knotificationplugin.h"
 #include "notifybypopup.h"
-#include "notifybyaudio.h"
+
 #include "notifybylogfile.h"
 #include "notifybytaskbar.h"
 #include "notifybyexecute.h"
+#include "debug_p.h"
+
+#ifdef HAVE_PHONON4QT5
+#include "notifybyaudio.h"
+#endif
+
+#ifdef HAVE_SPEECH
+#include "notifybytts.h"
+#endif
 
 typedef QHash<QString, QString> Dict;
 
 struct KNotificationManager::Private {
     QHash<int, KNotification *> notifications;
-    QHash<QString, KNotifyPlugin *> notifyPlugins;
+    QHash<QString, KNotificationPlugin *> notifyPlugins;
 
     // incremental ids for notifications
     int notifyIdCounter;
+    QStringList dirtyConfigCache;
 };
 
 class KNotificationManagerSingleton
@@ -66,30 +75,41 @@ KNotificationManager::KNotificationManager()
     qDeleteAll(d->notifyPlugins);
     d->notifyPlugins.clear();
     addPlugin(new NotifyByPopup(this));
-    //FIXME: port and reenable
-//     addPlugin(new NotifyBySound(this));
     addPlugin(new NotifyByExecute(this));
+    //FIXME: port and reenable
 //     addPlugin(new NotifyByLogfile(this));
+
+#ifdef HAVE_PHONON4QT5
     addPlugin(new NotifyByAudio(this));
-    //TODO reactivate on Mac/Win when KWindowSystem::demandAttention will implemented on this system.
-    #ifndef Q_WS_MAC
+#endif
+
     addPlugin(new NotifyByTaskbar(this));
-    #endif
-//     addPlugin(new NotifyByKTTS(this));
 
-    KService::List offers = KServiceTypeTrader::self()->query("KNotify/NotifyMethod");
+#ifdef HAVE_SPEECH
+    addPlugin(new NotifyByTTS(this));
+#endif
 
-    QVariantList args;
-    QString error;
+    QList<QObject*> plugins = KPluginLoader::instantiatePlugins(QStringLiteral("knotification/notifyplugins"),
+                                                                std::function<bool(const KPluginMetaData &)>(),
+                                                                this);
 
-    Q_FOREACH (const KService::Ptr service, offers) {
-        KNotifyPlugin *plugin = service->createInstance<KNotifyPlugin>(this, args, &error);
-        if (plugin) {
-            addPlugin(plugin);
+    Q_FOREACH (QObject *plugin, plugins) {
+        KNotificationPlugin *notifyPlugin = qobject_cast<KNotificationPlugin*>(plugin);
+        if (notifyPlugin) {
+            addPlugin(notifyPlugin);
         } else {
-            qDebug() << "Could not load plugin" << service->name() << "due to:" << error;
+            // not our/valid plugin, so delete the created object
+            plugin->deleteLater();
         }
     }
+
+    QDBusConnection::sessionBus().connect(QString(),
+                                          QStringLiteral("/Config"),
+                                          QStringLiteral("org.kde.knotification"),
+                                          QStringLiteral("reparseConfiguration"),
+                                          this,
+                                          SLOT(reparseConfiguration(QString)));
+
 }
 
 KNotificationManager::~KNotificationManager()
@@ -97,7 +117,7 @@ KNotificationManager::~KNotificationManager()
     delete d;
 }
 
-void KNotificationManager::addPlugin(KNotifyPlugin *notifyPlugin)
+void KNotificationManager::addPlugin(KNotificationPlugin *notifyPlugin)
 {
     d->notifyPlugins[notifyPlugin->optionName()] = notifyPlugin;
     connect(notifyPlugin, SIGNAL(finished(KNotification*)), this, SLOT(notifyPluginFinished(KNotification*)));
@@ -116,42 +136,66 @@ void KNotificationManager::notifyPluginFinished(KNotification *notification)
 void KNotificationManager::notificationActivated(int id, int action)
 {
     if (d->notifications.contains(id)) {
-        qDebug() << id << " " << action;
+        qCDebug(LOG_KNOTIFICATIONS) << id << " " << action;
         KNotification *n = d->notifications[id];
         n->activate(action);
         close(id);
     }
 }
 
-void KNotificationManager::notificationClosed(KNotification *notification)
+void KNotificationManager::notificationClosed()
 {
-    if (d->notifications.contains(notification->id())) {
-        qDebug() << notification->id();
-        KNotification *n = d->notifications[notification->id()];
-        d->notifications.remove(notification->id());
-        n->close();
+    KNotification *notification = qobject_cast<KNotification*>(sender());
+    if (!notification) {
+        return;
+    }
+    // We cannot do d->notifications.find(notification->id()); here because the
+    // notification->id() is -1 or -2 at this point, so we need to look for value
+    for (auto iter = d->notifications.begin(); iter != d->notifications.end(); ++iter) {
+        if (iter.value() == notification) {
+            d->notifications.erase(iter);
+            break;
+        }
     }
 }
 
 void KNotificationManager::close(int id, bool force)
 {
     if (force || d->notifications.contains(id)) {
-        KNotification *n = d->notifications.take(id);
-        qDebug() << "Closing notification" << id;
+        KNotification *n = d->notifications.value(id);
+        qCDebug(LOG_KNOTIFICATIONS) << "Closing notification" << id;
 
-        Q_FOREACH (KNotifyPlugin *plugin, d->notifyPlugins) {
-            plugin->close(n);
+        // Find plugins that are actually acting on this notification
+        // call close() only on those, otherwise each KNotificationPlugin::close()
+        // will call finish() which may close-and-delete the KNotification object
+        // before it finishes calling close on all the other plugins.
+        // For example: Action=Popup is a single actions but there is 5 loaded
+        // plugins, calling close() on the second would already close-and-delete
+        // the notification
+        KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
+        QString notifyActions = notifyConfig.readEntry(QStringLiteral("Action"));
+
+        Q_FOREACH (const QString &action, notifyActions.split('|')) {
+            if (!d->notifyPlugins.contains(action)) {
+                qCDebug(LOG_KNOTIFICATIONS) << "No plugin for action" << action;
+                continue;
+            }
+
+            d->notifyPlugins[action]->close(n);
         }
     }
 }
 
 int KNotificationManager::notify(KNotification *n)
 {
-    //FIXME: leaking?
-    KNotifyConfig *notifyConfig = new KNotifyConfig(n->appName(), n->contexts(), n->eventId());
+    KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
 
-    QString notifyActions = notifyConfig->readEntry("Action");
-    qDebug() << "Got notification \"" << n->eventId() <<"\" with actions:" << notifyActions;
+    if (d->dirtyConfigCache.contains(n->appName())) {
+        notifyConfig.reparseSingleConfiguration(n->appName());
+        d->dirtyConfigCache.removeOne(n->appName());
+    }
+
+    QString notifyActions = notifyConfig.readEntry(QStringLiteral("Action"));
 
     if (notifyActions.isEmpty() || notifyActions == QLatin1String("None")) {
         // this will cause KNotification closing itself fast
@@ -159,21 +203,22 @@ int KNotificationManager::notify(KNotification *n)
         return -1;
     }
 
-    d->notifications.insert(++d->notifyIdCounter, n);
+    d->notifications.insert(d->notifyIdCounter, n);
 
     Q_FOREACH (const QString &action, notifyActions.split('|')) {
         if (!d->notifyPlugins.contains(action)) {
-            qDebug() << "No plugin for action" << action;
+            qCDebug(LOG_KNOTIFICATIONS) << "No plugin for action" << action;
             continue;
         }
 
-        KNotifyPlugin *notifyPlugin = d->notifyPlugins[action];
+        KNotificationPlugin *notifyPlugin = d->notifyPlugins[action];
         n->ref();
-        qDebug() << "calling notify on" << notifyPlugin->optionName();
-        notifyPlugin->notify(n, notifyConfig);
+        qCDebug(LOG_KNOTIFICATIONS) << "Calling notify on" << notifyPlugin->optionName();
+        notifyPlugin->notify(n, &notifyConfig);
     }
 
-    return d->notifyIdCounter;
+    connect(n, &KNotification::closed, this, &KNotificationManager::notificationClosed);
+    return d->notifyIdCounter++;
 }
 
 void KNotificationManager::update(KNotification *n)
@@ -187,7 +232,7 @@ void KNotificationManager::update(KNotification *n)
 
     KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
 
-    Q_FOREACH (KNotifyPlugin *p, d->notifyPlugins) {
+    Q_FOREACH (KNotificationPlugin *p, d->notifyPlugins) {
         p->update(n, &notifyConfig);
     }
 }
@@ -206,4 +251,11 @@ void KNotificationManager::reemit(KNotification *n)
     notify(n);
 }
 
+void KNotificationManager::reparseConfiguration(const QString &app)
+{
+    if (!d->dirtyConfigCache.contains(app)) {
+        d->dirtyConfigCache << app;
+    }
+}
+
 #include "moc_knotificationmanager_p.cpp"
