@@ -20,19 +20,18 @@
 #include "kpassivepopup.h"
 
 #include <config-knotifications.h>
+#include "debug_p.h"
 
 // Qt
-#include <QApplication>
+#include <QGuiApplication>
 #include <QBitmap>
-#include <QDebug>
-#include <QDesktopWidget>
+#include <QScreen>
 #include <QLabel>
 #include <QLayout>
 #include <QBoxLayout>
 #include <QMouseEvent>
 #include <QPainter>
 #include <QPainterPath>
-#include <QDesktopWidget>
 #include <QPolygonF>
 #include <QStyle>
 #include <QTimer>
@@ -44,9 +43,11 @@
 #include <netwm.h>
 #endif
 
+#include <kwindowinfo.h>
+
 static const int DEFAULT_POPUP_TYPE = KPassivePopup::Boxed;
 static const int DEFAULT_POPUP_TIME = 6 * 1000;
-static const Qt::WindowFlags POPUP_FLAGS = Qt::Tool | Qt::X11BypassWindowManagerHint | Qt::WindowStaysOnTopHint | Qt::FramelessWindowHint;
+static const Qt::WindowFlags POPUP_FLAGS = Qt::Tool | Qt::WindowStaysOnTopHint | Qt::FramelessWindowHint;
 
 class KPassivePopup::Private
 {
@@ -61,7 +62,13 @@ public:
           hideTimer(new QTimer(q)),
           autoDelete(false)
     {
-        q->setWindowFlags(POPUP_FLAGS);
+#if HAVE_X11
+        if (QX11Info::isPlatformX11()) {
+            q->setWindowFlags(POPUP_FLAGS | Qt::X11BypassWindowManagerHint);
+        } else
+#else
+            q->setWindowFlags(POPUP_FLAGS);
+#endif
         q->setFrameStyle(QFrame::Box | QFrame::Plain);
         q->setLineWidth(2);
 
@@ -101,7 +108,7 @@ public:
     {
         // get screen-geometry for screen our anchor is on
         // (geometry can differ from screen to screen!
-        QRect deskRect = QApplication::desktop()->screenGeometry(anchor);
+        QRect deskRect = desktopRectForPoint(anchor);
 
         const int width = q->width();
         const int height = q->height();
@@ -190,7 +197,7 @@ public:
         int w = q->minimumSizeHint().width();
         int h = q->minimumSizeHint().height();
 
-        QRect r = QApplication::desktop()->screenGeometry(QPoint(x + w / 2, y + h / 2));
+        QRect r = desktopRectForPoint(QPoint(x + w / 2, y + h / 2));
 
         if (popupStyle == Balloon) {
             // find a point to anchor to
@@ -227,6 +234,19 @@ public:
 
         return QPoint(x, y);
     }
+
+    QRect desktopRectForPoint(const QPoint &point)
+    {
+        QList<QScreen*> screens = QGuiApplication::screens();
+        Q_FOREACH(const QScreen *screen, screens) {
+            if (screen->geometry().contains(point)) {
+                return screen->geometry();
+            }
+        }
+
+        // If no screen was found, return the primary screen's geometry
+        return QGuiApplication::primaryScreen()->geometry();
+    }
 };
 
 KPassivePopup::KPassivePopup(QWidget *parent, Qt::WindowFlags f)
@@ -279,7 +299,7 @@ void KPassivePopup::setView(QWidget *child)
 void KPassivePopup::setView(const QString &caption, const QString &text,
                             const QPixmap &icon)
 {
-    // qDebug() << "KPassivePopup::setView " << caption << ", " << text;
+    // qCDebug(LOG_KNOTIFICATIONS) << "KPassivePopup::setView " << caption << ", " << text;
     setView(standardView(caption, text, icon, this));
 }
 
@@ -295,7 +315,7 @@ QWidget *KPassivePopup::standardView(const QString &caption,
 
     QHBoxLayout *hb = 0;
     if (!icon.isNull()) {
-        hb = new QHBoxLayout(top);
+        hb = new QHBoxLayout;
         hb->setMargin(0);
         vb->addLayout(hb);
         d->ttlIcon = new QLabel(top);
@@ -347,7 +367,7 @@ int KPassivePopup::timeout() const
 
 void KPassivePopup::setTimeout(int delay)
 {
-    d->hideDelay = delay;
+    d->hideDelay = delay < 0 ? DEFAULT_POPUP_TIME : delay;
     if (d->hideTimer->isActive()) {
         if (delay) {
             d->hideTimer->start(delay);
@@ -425,7 +445,7 @@ void KPassivePopup::hideEvent(QHideEvent *)
 
 QPoint KPassivePopup::defaultLocation() const
 {
-    const QRect r = QApplication::desktop()->availableGeometry();
+    const QRect r = QGuiApplication::primaryScreen()->availableGeometry();
     return QPoint(r.left(), r.top());
 }
 
@@ -433,50 +453,37 @@ void KPassivePopup::positionSelf()
 {
     QRect target;
 
-    if (!d->window) {
-        target = QRect(defaultLocation(), QSize(0, 0));
+    if (d->window) {
 #if HAVE_X11
-    } else if (QX11Info::isPlatformX11()) {
-        NETWinInfo ni(QX11Info::connection(), d->window, QX11Info::appRootWindow(),
-                      NET::WMIconGeometry | NET::WMState);
+        if (QX11Info::isPlatformX11()) {
+            NETWinInfo ni(QX11Info::connection(), d->window, QX11Info::appRootWindow(),
+                          NET::WMIconGeometry | NET::WMState);
 
-        // Try to put the popup by the taskbar entry
-        if (!(ni.state() & NET::SkipTaskbar)) {
-            NETRect r = ni.iconGeometry();
-            target.setRect(r.pos.x, r.pos.y, r.size.width, r.size.height);
-        }
-        // If that failed, put it by the window itself
-        if (target.isNull()) {
-            NETRect dummy;
-            NETRect r;
-            ni.kdeGeometry(dummy, r);
-            target.setRect(r.pos.x, r.pos.y,
-                           r.size.width, r.size.height);
-        }
-#else
-    } else {
-        // Our only choice is to put it by the window
-
-        // Easy case: we were passed one of our own widgets
-        QWidget *widget = QWidget::find(d->winId);
-        if (widget) {
-            target = widget->geometry();
-            if (target.isNull()) {
-                const QRect d = QApplication::desktop()->availableGeometry(widget);
-                target.setRect(d.x(), d.y(), 0, 0);
-            }
-        } else {
-            // NB: this will not work on X11 (hence the fallback in the NET
-            //     code above) as the XCB platform plugin does not collect
-            //     information about foreign windows
-            QWindow *window = QWindow::fromWinId(d->winId);
-            target = window->geometry();
-            if (target.isNull()) {
-                const QRect d = window->screen()->availableGeometry();
-                target.setRect(d.x(), d.y(), 0, 0);
+            // Try to put the popup by the taskbar entry
+            if (!(ni.state() & NET::SkipTaskbar)) {
+                NETRect r = ni.iconGeometry();
+                target.setRect(r.pos.x, r.pos.y, r.size.width, r.size.height);
             }
         }
 #endif
+        // If that failed, put it by the window itself
+
+        if (target.isNull()) {
+            // Avoid making calls to the window system if we can
+            QWidget *widget = QWidget::find(d->window);
+            if (widget) {
+                target = widget->geometry();
+            }
+        }
+        if (target.isNull()) {
+            KWindowInfo info(d->window, NET::WMGeometry);
+            if (info.valid()) {
+                target = info.geometry();
+            }
+        }
+    }
+    if (target.isNull()) {
+        target = QRect(defaultLocation(), QSize(0, 0));
     }
     moveNear(target);
 }
@@ -564,7 +571,7 @@ KPassivePopup *KPassivePopup::message(int popupStyle, const QString &caption, co
     pop->setPopupStyle(popupStyle);
     pop->setAutoDelete(true);
     pop->setView(caption, text, icon);
-    pop->d->hideDelay = timeout;
+    pop->d->hideDelay = timeout < 0 ? DEFAULT_POPUP_TIME : timeout;
     if (p.isNull()) {
         pop->show();
     } else {
@@ -592,7 +599,7 @@ KPassivePopup *KPassivePopup::message(int popupStyle, const QString &caption, co
     pop->setPopupStyle(popupStyle);
     pop->setAutoDelete(true);
     pop->setView(caption, text, icon);
-    pop->d->hideDelay = timeout;
+    pop->d->hideDelay = timeout < 0 ? DEFAULT_POPUP_TIME : timeout;
     if (p.isNull()) {
         pop->show();
     } else {
@@ -609,7 +616,7 @@ KPassivePopup *KPassivePopup::message(int popupStyle, const QString &caption, co
     pop->setPopupStyle(popupStyle);
     pop->setAutoDelete(true);
     pop->setView(caption, text, icon);
-    pop->d->hideDelay = timeout;
+    pop->d->hideDelay = timeout < 0 ? DEFAULT_POPUP_TIME : timeout;
     QPoint pos = pop->d->calculateNearbyPoint(parent->geometry());
     pop->show(pos);
     pop->moveNear(parent->geometry());
